import unittest
from typing import cast

from dataladmetadatamodel.common import get_top_level_metadata_objects
from dataladmetadatamodel.datasettree import DatasetTree
from dataladmetadatamodel.metadata import Metadata
from dataladmetadatamodel.metadatapath import MetadataPath


class TestRemote(unittest.TestCase):

    def test_basic_remote(self):
        tree_version_list, uuid_set = get_top_level_metadata_objects(
            "git",
            "https://github.com/datalad/test_metadata"
        )
        self.assertEqual(len(tree_version_list.version_set), 1)
        self.assertEqual(len(uuid_set.uuid_set), 1)

        for version, element_info in tree_version_list.versioned_elements:
            time_stamp, dataset_path, dataset_tree = element_info
            dataset_tree = cast(DatasetTree, dataset_tree)
            dataset_tree.read_in()
            dataset_paths = dataset_tree.dataset_paths
            self.assertEqual(27, len(dataset_paths))

            mrr = dataset_tree.get_metadata_root_record(MetadataPath("study-104"))
            file_tree = mrr.file_tree
            file_paths = list(file_tree.get_paths_recursive())
            self.assertEqual(7, len(file_paths))

            file_tree.add_metadata(MetadataPath("a/b/c"), Metadata())
            self.assertRaises(RuntimeError, file_tree.write_out)



import unittest
from unittest.mock import patch
from pathlib import (
    PurePosixPath,
    PureWindowsPath
)

from dataladmetadatamodel.metadatapath import MetadataPath


class TestMetadataPath(unittest.TestCase):
    def test_empty_join(self):
        self.assertEqual(
            MetadataPath("") / MetadataPath(""),
            MetadataPath(""))

    def test_empty_leading(self):
        self.assertEqual(
            MetadataPath("") / MetadataPath("") / MetadataPath("a"),
            MetadataPath("a"))

    def test_common(self):
        self.assertEqual(
            MetadataPath("") / MetadataPath("") / MetadataPath(
                "a") / MetadataPath("b"),
            MetadataPath("a/b"))

    def test_multiple_dashes(self):
        self.assertEqual(
            MetadataPath("/") / MetadataPath("a//") / MetadataPath("b"),
            MetadataPath("a/b"))

    def test_intermediate_root(self):
        self.assertEqual(
            MetadataPath("/") / MetadataPath("a//") / MetadataPath("/b"),
            MetadataPath("a/b"))

    def test_windows_paths(self):
        # Enforce windows prefix_path interpretation
        with patch("dataladmetadatamodel.metadatapath.PurePosixPath", new=PureWindowsPath):
            metadata_path = MetadataPath("a\\b\\c")
        self.assertEqual(metadata_path, MetadataPath("a/b/c"))

        with patch("dataladmetadatamodel.metadatapath.PurePosixPath", new=PureWindowsPath):
            metadata_path = MetadataPath("a/b/c")
        self.assertEqual(metadata_path, MetadataPath("a/b/c"))

    def test_posix_paths(self):
        with patch("dataladmetadatamodel.metadatapath.PurePosixPath", new=PurePosixPath):
            metadata_path = MetadataPath("a/b/c")
        self.assertEqual(metadata_path, MetadataPath("a/b/c"))


if __name__ == '__main__':
    unittest.main()



"""
Different remote repositories have to be cached in different
local repositories to avoid conflicts in reference name space
"""
import hashlib
from pathlib import Path

import appdirs

from .gitbackend.subprocess import (
    git_fetch_object,
    git_fetch_reference,
    git_init,
    git_object_exists_locally,
)


cache_repository_base = (
    Path(appdirs.user_cache_dir("datalad-metalad"))
    / "local-git-cache"
)


def get_name_hash(name: str) -> str:
    return hashlib.sha256(name.encode()).hexdigest()


def get_repo_cache_path(remote_repo: str) -> Path:
    return cache_repository_base / get_name_hash(remote_repo)


def ensure_cache_exists(remote_repo: str):
    cache_path = get_repo_cache_path(remote_repo)
    if not cache_path.exists():
        git_init(str(cache_path))


def cache_object(remote_repo: str, object_id: str):

    ensure_cache_exists(remote_repo)
    cache_repo = str(get_repo_cache_path(remote_repo))

    if object_id.startswith("refs"):
        git_fetch_reference(
            cache_repo,
            remote_repo,
            object_id,
            object_id)
    elif not git_object_exists_locally(cache_repo, object_id):
        git_fetch_object(
            cache_repo,
            remote_repo,
            object_id)


def get_cache_realm(remote_repo: str) -> Path:
    return get_repo_cache_path(remote_repo)



from typing import Dict

from dataladmetadatamodel.mapper.gitmapper.objectreference import (
    GitReference,
    add_blob_reference,
)
from dataladmetadatamodel.mapper.gitmapper.gitbackend.subprocess import (
    git_load_str,
    git_save_str,
)
from dataladmetadatamodel.mapper.mapper import Mapper
from dataladmetadatamodel.mapper.reference import Reference

from .gitblobcache import GitBlobCache


class MetadataGitMapper(Mapper):

    metadata_caches: Dict[str, GitBlobCache] = dict()

    @classmethod
    def get_cache(cls, realm: str):
        return cls.metadata_caches.get(realm, None)

    @classmethod
    def cache_realm(cls, realm: str):
        if cls.get_cache(realm) is not None:
            raise RuntimeError(f"already caching realm: {realm}")
        cls.metadata_caches[realm] = GitBlobCache(realm)

    @classmethod
    def flush_realm(cls, realm: str):
        cache = cls.get_cache(realm)
        if cache is None:
            raise RuntimeError(f"realm is not cached: {realm}")
        cache.flush()
        del cls.metadata_caches[realm]

    def map_in_impl(self,
                    metadata: "Metadata",
                    realm: str,
                    meta_reference: Reference) -> None:

        from dataladmetadatamodel.metadata import Metadata
        from dataladmetadatamodel.mapper.reference import Reference

        assert isinstance(metadata, Metadata)

        reference = Reference.from_json_str(
            git_load_str(realm, meta_reference.location))

        metadata.init_from_json(
            git_load_str(realm, reference.location))

    def map_out_impl(self,
                     metadata: "Metadata",
                     realm: str,
                     force_write: bool) -> Reference:

        from dataladmetadatamodel.metadata import Metadata
        assert isinstance(metadata, Metadata)

        cache = self.get_cache(realm)
        if cache is None:
            return self.map_out_impl_immediate(metadata, realm, force_write)
        else:
            return self.map_out_impl_cached(cache, metadata, realm)

    def map_out_impl_immediate(self,
                               metadata: "Metadata",
                               realm: str,
                               force_write: bool) -> Reference:

        # Save metadata object and add it to the
        # blob-references. That is necessary because
        # metadata blobs are only referenced in
        # persisted Reference-objects, i.e. in
        # JSON-strings that are stored in the
        # repository.
        metadata_blob_location = git_save_str(realm, metadata.to_json())
        add_blob_reference(metadata_blob_location)

        # Save reference. NB we don't have to save
        # metadata_reference_blob_location to the
        # reference objects because they will be
        # reachable by a git tree entry.
        metadata_reference_blob_location = git_save_str(
            realm,
            Reference(
                "Metadata",
                metadata_blob_location).to_json_str())

        # Return reference to reference, although it specifies
        # the class Metadata, it points actually to a
        # persisted Reference object. But this is only known
        # internally in this mapper.
        return Reference(
            "Metadata",
            metadata_reference_blob_location)

    def map_out_impl_cached(self,
                            cache: GitBlobCache,
                            metadata: "Metadata",
                            realm: str) -> Reference:

        # Cache metadata, and cache a reference object. the
        # following code is "cache"-equivalent to the immediate case
        metadata_blob_location = cache.cache_blob(realm, metadata.to_json())
        add_blob_reference(metadata_blob_location)

        metadata_reference_blob_location = cache.cache_blob(
            realm,
            Reference(
                "Metadata",
                metadata_blob_location).to_json_str())

        return Reference(
            "Metadata",
            metadata_reference_blob_location)



from typing import (
    Iterable,
    List,
    Optional
)
import unittest

from dataladmetadatamodel.modifiableobject import ModifiableObject


destination = "/tmp/t"


class SUTModifiableObject(ModifiableObject):
    def __init__(self, saved_on: Optional[str] = None):
        super().__init__()
        if saved_on is not None:
            self.set_saved_on(saved_on)
        self.something = "Something " * 100

    @property
    def modifiable_sub_objects(self) -> Iterable:
        return []


class TestModifiableObject(unittest.TestCase):

    def test_new_clean(self):
        mo = SUTModifiableObject()
        self.assertFalse(mo.is_saved_on(destination))

    def test_touching_cleaning(self):
        mo = SUTModifiableObject()
        self.assertFalse(mo.is_saved_on(destination))
        mo.set_saved_on(destination)
        self.assertTrue(mo.is_saved_on(destination))

    def test_sub_object_modification(self):
        class Bag(ModifiableObject):
            def __init__(self, _sub_objects: List[ModifiableObject]):
                super().__init__()
                self.sub_objects = _sub_objects

            @property
            def modifiable_sub_objects(self) -> Iterable[ModifiableObject]:
                return self.sub_objects

        sub_objects = [SUTModifiableObject(destination) for _ in range(3)]

        bag = Bag(sub_objects)
        bag.set_saved_on(destination)

        self.assertTrue(bag.is_saved_on(destination), f"Expected complete bag is saved on {destination}")

        sub_objects[0].touch()
        self.assertFalse(bag.is_saved_on(destination), f"Expected modified bag is not saved on {destination}")


if __name__ == '__main__':
    unittest.main()



"""
Simple lock interface.
"""
import os
import subprocess
import time
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import Dict

from fasteners import InterProcessLock

from dataladmetadatamodel.log import logger


PID = os.getpid()

GIT_MAPPER_LOCK_FILE_NAME = "datalad-metalad-write.lock"
read_write_locked = dict()


@dataclass
class LockState:
    counter: int
    lock: InterProcessLock


def _get_lock_state(lock_dict: dict, lock_dir: Path) -> LockState:
    if lock_dir not in lock_dict:
        lock_dict[lock_dir] = LockState(
            0,
            InterProcessLock(str(lock_dir / GIT_MAPPER_LOCK_FILE_NAME)))
    return lock_dict[lock_dir]


@contextmanager
def locked_backend(realm: Path):
    """ lock and unlock the backend given by realm """
    lock_backend(realm)
    try:
        yield None
    finally:
        unlock_backend(realm)


def lock_backend(realm: Path):
    lock_dir = get_lock_dir(realm)
    lock_state = _get_lock_state(read_write_locked, lock_dir)
    if lock_state.counter == 0:
        lock_time = time.time()
        lock_state.lock.acquire()
        lock_time = time.time() - lock_time
        logger.debug(
            "process {} locked git backend using file {} in {} seconds".format(
                PID,
                lock_dir,
                int(lock_time)))
    lock_state.counter += 1


def unlock_backend(realm: Path):
    lock_dir = get_lock_dir(realm, create_directory=False)
    assert lock_dir.exists()
    lock_state = _get_lock_state(read_write_locked, lock_dir)
    assert lock_state.counter > 0
    lock_state.counter -= 1
    if lock_state.counter == 0:
        logger.debug(
            "process {} unlocks git backend using file {}".format(PID, lock_dir))
        lock_state.lock.release()


def get_lock_dir(realm: Path, create_directory: bool = True) -> Path:
    lock_dir = realm / ".git"
    if not lock_dir.exists() and create_directory:
        # if exist_ok is False, we might get an error if
        # a concurrent mapper tries to lock the same
        # dataset.
        os.makedirs(lock_dir, exist_ok=True)
    return lock_dir


def create_file_tree(location: Path, content: Dict):
    for name, value_or_dict in content.items():
        new_location = location / name
        if isinstance(value_or_dict, str):
            new_location.write_text(value_or_dict)
        else:
            new_location.mkdir()
            create_file_tree(new_location, value_or_dict)


def create_git_repo(location: Path, content: Dict):
    subprocess.run(["git", "init", str(location)], check=True)
    create_file_tree(location, content)
    subprocess.run(
        ["git", "-C", str(location), "add", "."],
        check=True
    )

    subprocess.run(
        ["git", "-C", str(location), "commit", "-m", "create repo"],
        check=True
    )

    commit = subprocess.run(
        ["git", "-C", str(location), "cat-file", "-p", "HEAD"],
        check=True,
        stdout=subprocess.PIPE
    )
    return commit.stdout.decode().splitlines()[0].split()[1]



from typing import (
    cast,
    Dict,
    Generator,
    Iterable,
    Optional,
    Tuple,
    Union,
)

from dataladmetadatamodel.datasettree import DatasetTree
from dataladmetadatamodel.mappableobject import MappableObject
from dataladmetadatamodel.metadatapath import MetadataPath
from dataladmetadatamodel.metadatarootrecord import MetadataRootRecord
from dataladmetadatamodel.mapper.reference import Reference


unknown_version = "unknown"


class VersionRecord:
    def __init__(self,
                 time_stamp: str,
                 prefix_path: Optional[MetadataPath],
                 element: Union[DatasetTree, MetadataRootRecord]):

        assert isinstance(time_stamp, str)
        assert isinstance(prefix_path, (type(None), MetadataPath)), f"prefix_path is type: {type(prefix_path).__name__}"
        assert isinstance(element, (DatasetTree, MetadataRootRecord))

        self.time_stamp = time_stamp
        self.prefix_path = prefix_path or MetadataPath("")
        self.element = element

    def deepcopy(self,
                 new_mapper_family: Optional[str] = None,
                 new_destination: Optional[str] = None
                 ) -> "VersionRecord":

        return VersionRecord(
            self.time_stamp,
            self.prefix_path,
            self.element.deepcopy(
                new_mapper_family,
                new_destination))


class VersionList(MappableObject):
    def __init__(self,
                 initial_set: Optional[Dict[str, Dict[MetadataPath, VersionRecord]]] = None,
                 realm: Optional[str] = None,
                 reference: Optional[Reference] = None):

        assert isinstance(initial_set, (type(None), dict))
        assert isinstance(realm, (type(None), str))
        assert isinstance(reference, (type(None), Reference))

        super().__init__(realm, reference)
        self.version_set: Dict[str, Dict[MetadataPath, VersionRecord]] = \
            initial_set or dict()

    def __len__(self):
        return sum([len(path_dict) for path_dict in self.version_set.values()])

    def purge_impl(self):
        for version, prefixed_set in self.version_set.items():
            for version_record in prefixed_set.values():
                version_record.element.purge()
        self.version_set = dict()

    def _get_version_record(self,
                            primary_data_version: str,
                            prefix_path: MetadataPath) -> VersionRecord:
        self.ensure_mapped()
        return self.version_set[primary_data_version][prefix_path]

    def _get_version_records(self,
                             primary_data_version: str
                             ) -> Iterable[VersionRecord]:
        self.ensure_mapped()
        return self.version_set[primary_data_version].values()

    def _iterate_version_records(self) -> Generator:
        self.ensure_mapped()
        yield from (
            (version, version_record)
            for version, prefixed_set in self.version_set.items()
            for version_record in prefixed_set.values()
        )

    def modifiable_sub_objects_impl(self) -> Iterable[MappableObject]:
        yield from map(
            lambda vvr: vvr[1].element,
            self._iterate_version_records())

    def versions(self) -> Iterable:
        self.ensure_mapped()
        yield from map(
            lambda vvr: vvr[0],
            self._iterate_version_records())

    def versions_and_prefix_paths(self) -> Iterable:
        self.ensure_mapped()
        yield from map(
            lambda vvr: (vvr[0], vvr[1].prefix_path),
            self._iterate_version_records())

    def get_versioned_element(self,
                              primary_data_version: str,
                              prefix_path: MetadataPath,
                              ) -> Tuple[str, MetadataPath, MappableObject]:
        """
        Get the dataset tree or metadata root record,
        its timestamp and prefix_path for the given version.
        If it is not mapped yet, it will be mapped.
        """
        self.ensure_mapped()
        version_record = self._get_version_record(
            primary_data_version,
            prefix_path)
        version_record.element.read_in()
        return (
            version_record.time_stamp,
            version_record.prefix_path,
            version_record.element)

    def get_versioned_elements(self,
                               primary_data_version: str,
                               ) -> Iterable[Tuple[str, MetadataPath, MappableObject]]:
        """
        Get all dataset trees or metadata root records, including their
        timestamp and prefix paths for the given version.
        If they are not mapped yet, they will be mapped.
        """
        self.ensure_mapped()
        for version_record in self._get_version_records(primary_data_version):
            version_record.element.read_in()
            yield (
                version_record.time_stamp,
                version_record.prefix_path,
                version_record.element)

    @property
    def versioned_elements(self
                           ) -> Iterable[Tuple[str, Tuple[str, MetadataPath, MappableObject]]]:
        """
        Get an iterable of all versions and their records
        """
        self.ensure_mapped()
        yield from (
            (
                version,
                (
                    version_record.time_stamp,
                    version_record.prefix_path,
                    version_record.element
                )
            )
            for version, version_record in self._iterate_version_records())

    def set_versioned_element(self,
                              primary_data_version: str,
                              time_stamp: str,
                              prefix_path: MetadataPath,
                              element: Union[DatasetTree, MetadataRootRecord]):
        """
        Set a new or updated dataset tree.
        Existing references are deleted.
        The entry is marked as dirty.
        """
        self.ensure_mapped()
        self.touch()
        if primary_data_version not in self.version_set:
            self.version_set[primary_data_version] = dict()
        self.version_set[primary_data_version][prefix_path] = VersionRecord(
            time_stamp,
            prefix_path,
            element)

    def unget_versioned_element(self,
                                primary_data_version: str,
                                prefix_path: MetadataPath,
                                new_destination: Optional[str] = None):
        """
        Remove a metadata record from memory. First, persist the
        current status via save_object ---which will write it to
        the backend, if it was changed or--- then purge it
        from memory.
        """
        version_record = self.version_set[primary_data_version][prefix_path]
        version_record.element.write_out(new_destination)
        version_record.element.purge()

    def deepcopy_impl(self,
                      new_mapper_family: Optional[str] = None,
                      new_destination: Optional[str] = None,
                      **kwargs
                      ) -> "VersionList":

        path_prefix = kwargs.get("path_prefix", MetadataPath(""))

        copied_version_list = VersionList()
        for primary_data_version, prefix_set in self.version_set.items():
            for version_record in prefix_set.values():
                copied_version_list.set_versioned_element(
                    primary_data_version,
                    version_record.time_stamp,
                    path_prefix / version_record.prefix_path,
                    version_record.element.deepcopy(
                        new_mapper_family,
                        new_destination))

        copied_version_list.write_out(new_destination)
        copied_version_list.purge()
        return copied_version_list


class TreeVersionList(VersionList):
    """
    Thin wrapper around version list to support tree-version list
    specific mapping, for example in the git mapper, which will
    update a reference, if mapping an TreeVersionList instance.
    """

    def get_dataset_tree(self,
                         primary_data_version: str,
                         prefix_path: MetadataPath,
                         ) -> Tuple[str, MetadataPath, DatasetTree]:

        self.ensure_mapped()
        time_stamp, prefix_path, dataset_tree = super().get_versioned_element(
            primary_data_version,
            prefix_path)
        dataset_tree = cast(DatasetTree, dataset_tree)
        return time_stamp, prefix_path, dataset_tree

    def get_dataset_trees(self,
                          primary_data_version: str,
                          ) -> Iterable[Tuple[str, MetadataPath, DatasetTree]]:

        self.ensure_mapped()
        return super().get_versioned_elements(primary_data_version)

    def set_dataset_tree(self,
                         primary_data_version: str,
                         time_stamp: str,
                         prefix_path: MetadataPath,
                         dataset_tree: DatasetTree,
                         ):

        self.ensure_mapped()
        self.touch()
        return super().set_versioned_element(
            primary_data_version=primary_data_version,
            time_stamp=time_stamp,
            prefix_path=prefix_path,
            element=dataset_tree)

    def unget_dataset_tree(self,
                           primary_data_version: str,
                           prefix_path: MetadataPath,
                           new_destination: str):

        assert isinstance(primary_data_version, str)
        assert isinstance(new_destination, str)
        super().unget_versioned_element(
            primary_data_version,
            prefix_path,
            new_destination)

    def deepcopy_impl(self,
                      new_mapper_family: Optional[str] = None,
                      new_destination: Optional[str] = None,
                      **kwargs
                      ) -> "TreeVersionList":

        path_prefix = kwargs.get("path_prefix", MetadataPath(""))

        copied_version_list = TreeVersionList()
        for primary_data_version, prefix_set in self.version_set.items():
            for version_record in prefix_set.values():
                copied_version_list.set_versioned_element(
                    primary_data_version,
                    version_record.time_stamp,
                    path_prefix / version_record.prefix_path,
                    cast(MetadataRootRecord,
                         version_record.element.deepcopy(
                            new_mapper_family,
                            new_destination)))

        copied_version_list.write_out(new_destination)
        copied_version_list.purge()
        return copied_version_list



import unittest
from unittest import mock

from dataladmetadatamodel.mapper.gitmapper.gitblobcache import hash_blob
from dataladmetadatamodel.metadata import Metadata
from dataladmetadatamodel.metadatapath import MetadataPath
from dataladmetadatamodel.metadatarootrecord import MetadataRootRecord
from dataladmetadatamodel.tests.utils import (
    create_file_tree,
    get_location,
    get_uuid,
)
from dataladmetadatamodel.versionlist import (
    VersionList,
    VersionRecord,
)


test_realm_name = "ewkd0iasd"


class TestVersionListMapper(unittest.TestCase):

    def test_basic_unmapping(self):

        with mock.patch("dataladmetadatamodel.mapper.gitmapper."
                        "versionlistmapper.git_save_json") as save_json, \
             mock.patch("dataladmetadatamodel.mapper.gitmapper."
                        "metadatamapper.git_save_str") as save_str, \
             mock.patch("dataladmetadatamodel.mapper.gitmapper."
                        "versionlistmapper.git_update_ref"):

            save_json.configure_mock(return_value=get_location(0))
            save_str.configure_mock(return_value=get_location(1))

            version_list = VersionList()
            version_list.write_out("/tmp/t1")
            representation = save_json.call_args[0][1]
            self.assertEqual(representation, [])

    def test_complex_unmapping(self):

        with mock.patch("dataladmetadatamodel.mapper.gitmapper."
                        "versionlistmapper.git_save_json") as save_json, \
             mock.patch("dataladmetadatamodel.mapper.gitmapper."
                        "metadatamapper.git_save_str") as save_str, \
             mock.patch("dataladmetadatamodel.mapper.gitmapper."
                        "metadatarootrecordmapper.git_save_json") as save_json_2, \
             mock.patch("dataladmetadatamodel.mapper.gitmapper."
                        "mtreenodemapper.git_save_tree_node") as save_tree_node, \
             mock.patch("dataladmetadatamodel.mapper.gitmapper."
                        "gitblobcache.git_save_file_list") as file_list_save, \
             mock.patch("dataladmetadatamodel.mapper.gitmapper."
                        "versionlistmapper.git_update_ref"):

            save_json.return_value = get_location(0)
            save_str.return_value = get_location(1)
            save_json_2.return_value = get_location(3)
            save_tree_node.return_value = get_location(4)
            file_list_save.side_effect = lambda r, l: [
                hash_blob(open(e, "rb").read())
                for e in l
            ]

            version_list = VersionList(initial_set={
                "v0": {
                    MetadataPath("subset0"): VersionRecord(
                        str("1.1"),
                        MetadataPath("subset0"),
                        MetadataRootRecord(
                            get_uuid(0),
                            "version1",
                            Metadata(),
                            create_file_tree(
                                [MetadataPath("a/b"),
                                 MetadataPath("d/e")]
                            )
                        )
                    )
                }
            })

            version_list.write_out("/tmp/t1")
            representation = save_json.call_args[0][1]
            self.assertEqual(
                representation,
                [{
                    'primary_data_version': 'v0',
                    'time_stamp': '1.1',
                    'path': 'subset0',
                    'dataset_tree': {
                        '@': {
                            'type': 'Reference',
                            'version': '2.0'
                        },
                        'class_name': 'MetadataRootRecord',
                        'location': get_location(3)
                    }
                }]
            )


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest import mock
from uuid import UUID

from dataladmetadatamodel.mapper.gitmapper.gitblobcache import hash_blob
from dataladmetadatamodel.metadata import Metadata
from dataladmetadatamodel.metadatapath import MetadataPath
from dataladmetadatamodel.metadatarootrecord import MetadataRootRecord
from dataladmetadatamodel.tests.utils import create_file_tree_with_metadata

from .... import version_string


uuid_0 = UUID("00000000000000000000000000000000")
dataset_version = "000000011111222223333"


location_0 = "a000000000000000000000000000000000000000"
location_1 = "a000000000000000000000000000000000000001"
location_2 = "a000000000000000000000000000000000000002"
location_3 = "a000000000000000000000000000000000000003"
location_4 = "a000000000000000000000000000000000000004"

default_paths = [
    MetadataPath("a/b/c"),
    MetadataPath("a/b/a"),
    MetadataPath("b"),
    MetadataPath("c/d/e"),
    MetadataPath("a/x")
]


class TestMetadataMapper(unittest.TestCase):

    def test_basic_unmapping(self):

        file_tree = create_file_tree_with_metadata(default_paths, [
            Metadata()
            for _ in default_paths])

        mrr = MetadataRootRecord(
            uuid_0,
            dataset_version,
            Metadata(),
            file_tree)

        with mock.patch("dataladmetadatamodel.mapper.gitmapper."
                        "metadatarootrecordmapper.git_save_json") as save, \
             mock.patch("dataladmetadatamodel.mapper.gitmapper."
                        "metadatamapper.git_save_str") as str_save, \
             mock.patch("dataladmetadatamodel.mapper.gitmapper."
                        "gitblobcache.git_save_file_list") as file_list_save, \
             mock.patch("dataladmetadatamodel.mapper.gitmapper."
                        "mtreenodemapper.git_save_tree_node") as save_tree_node:

            save.configure_mock(return_value=location_0)
            str_save.configure_mock(return_value=location_1)
            save_tree_node.configure_mock(return_value=location_3)
            file_list_save.side_effect = lambda r, l: [
                hash_blob(open(e, "rb").read())
                for e in l
            ]

            reference = mrr.write_out("/tmp/t1", "git")
            self.assertEqual(reference.location, location_0)

            save.assert_called_once()
            representation = save.call_args[0][1]
            self.assertEqual(
                representation,
                {
                    'dataset_identifier': str(uuid_0),
                    'dataset_version': dataset_version,
                    'dataset_level_metadata': {
                        '@': {
                            'type': 'Reference',
                            'version': version_string
                        },
                        'class_name': 'Metadata',
                        'location': location_1
                    },
                    'file_tree': {
                        '@': {
                            'type': 'Reference',
                            'version': version_string
                        },
                        'class_name': 'MTreeNode',
                        'location': location_3}
                }
            )


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest import mock
from uuid import UUID


location_0 = "a000000000000000000000000000000000000000"
location_1 = "a000000000000000000000000000000000000001"

uuid_0 = UUID("00000000000000000000000000000000")
uuid_1 = UUID("00000000000000000000000000000001")


expected_tree_args = {
    ('100644', 'blob', 'a000000000000000000000000000000000000001', '00000000-0000-0000-0000-000000000001'),
    ('100644', 'blob', 'a000000000000000000000000000000000000001', '00000000-0000-0000-0000-000000000000')
}


class TestUUIDSetMapper(unittest.TestCase):

    def test_basic_unmapping(self):
        from dataladmetadatamodel.uuidset import UUIDSet
        from dataladmetadatamodel.versionlist import VersionList

        with mock.patch("dataladmetadatamodel.mapper.gitmapper.uuidsetmapper.git_save_tree") as save_tree, \
                mock.patch("dataladmetadatamodel.mapper.gitmapper.uuidsetmapper.git_update_ref") as update_ref, \
                mock.patch("dataladmetadatamodel.mapper.gitmapper.versionlistmapper.git_save_json") as save_json:

            save_tree.configure_mock(return_value=location_0)
            save_json.configure_mock(return_value=location_1)

            uuid_set = UUIDSet({
                uuid_0: VersionList(),
                uuid_1: VersionList()
            })

            uuid_set.write_out("/tmp/t1")

            tree_args = save_tree.call_args[0][1]
            self.assertEqual(tree_args, expected_tree_args)


if __name__ == '__main__':
    unittest.main()



"""
A representation of metadata sources.

A MetadataSource objects makes the underlying
storage system visible. E.g. a MetadataSource
object for git-stored metadata, supports the
implementation to read, write, and copy
this data from, to, and between git
repositories. (This is in contrast to
the metadata model classes, that never
deals with a backend, e.g. a git-repository,
for their persistence, instead they are
mapped by mappers onto a specific backend.)
"""
import abc
import json
import logging
import subprocess
from copy import deepcopy
from pathlib import Path
from typing import IO, Optional

from dataladmetadatamodel import (
    JSONObject,
    check_serialized_version,
    version_string
)


logger = logging.getLogger("datalad.metadata.model")


class MetadataSource(abc.ABC):
    """
    Base class for all metadata source handlers
    """

    TYPE_KEY = "metadata_source_type"

    @abc.abstractmethod
    def write_object_to(self, file_descriptor):
        raise NotImplementedError

    @abc.abstractmethod
    def copy_object_to(self, destination: Path):
        raise NotImplementedError

    @abc.abstractmethod
    def to_json_obj(self) -> JSONObject:
        raise NotImplementedError

    def to_json_str(self) -> str:
        return json.dumps(self.to_json_obj())

    @staticmethod
    def from_json_obj(json_obj: JSONObject) -> Optional["MetadataSource"]:
        source_type = json_obj.get(MetadataSource.TYPE_KEY, None)
        if source_type is None:
            logger.error(
                f"key `{MetadataSource.TYPE_KEY}´ not found in "
                f"json object: {json.dumps(json_obj)}.")
            return None
        if source_type == LocalGitMetadataSource.TYPE:
            return LocalGitMetadataSource.from_json_obj(json_obj)
        elif source_type == ImmediateMetadataSource.TYPE:
            return ImmediateMetadataSource.from_json_obj(json_obj)
        else:
            logger.error(f"unknown metadata source type: `{source_type}´")
            return None

    @staticmethod
    def from_json_str(json_string: str) -> Optional["MetadataSource"]:
        return MetadataSource.from_json_obj(json.loads(json_string))


class LocalGitMetadataSource(MetadataSource):

    TYPE = "LocalGitMetadataSource"

    def __init__(self,
                 git_repository_path: Path,
                 object_reference: str
                 ):

        assert isinstance(object_reference, str)
        super().__init__()
        self.git_repository_path = git_repository_path
        self.object_reference = object_reference.strip()

    def __eq__(self, other):
        return (
            isinstance(other, LocalGitMetadataSource)
            and self.git_repository_path == other.git_repository_path
            and self.object_reference == other.object_reference)

    def write_object_to(self, file_descriptor: IO):
        command = f"git --git-dir {self.git_repository_path / '.git'} " \
                  f"cat-file blob {self.object_reference}"
        result = subprocess.run(command, stdout=file_descriptor, shell=True)
        if result.returncode != 0:
            raise RuntimeError(
                f"subprocess returned {result.returncode}, "
                f"command: {command}")

    def copy_object_to(self, destination_repository: Path) -> str:
        """
        copy an object from the LocalGitMetadataSource
        instance into the git repository given by
        destination_repository.
        """
        command = f"git --git-dir {self.git_repository_path / '.git'} " \
                  f"cat-file blob {self.object_reference}|" \
                  f"git --git-dir {destination_repository / '.git'} " \
                  f"hash-object -w --stdin"
        copied_object_reference = subprocess.check_output(
            command,
            shell=True).decode().strip()
        assert copied_object_reference == self.object_reference
        return copied_object_reference

    def to_json_obj(self) -> JSONObject:
        return {
            "@": dict(
                type="LocalGitMetadataSource",
                version=version_string
            ),
            MetadataSource.TYPE_KEY: LocalGitMetadataSource.TYPE,
            "git_repository_path": self.git_repository_path.as_posix(),
            "object_reference": self.object_reference}

    @staticmethod
    def from_json_obj(json_obj: JSONObject) -> Optional["LocalGitMetadataSource"]:
        try:
            assert json_obj["@"]["type"] == "LocalGitMetadataSource"
            check_serialized_version(json_obj)
            assert json_obj[MetadataSource.TYPE_KEY] == LocalGitMetadataSource.TYPE
            return LocalGitMetadataSource(
                Path(json_obj["git_repository_path"]),
                json_obj["object_reference"])
        except KeyError as key_error:
            logger.error(
                f"could not read LocalGitMetadataSource from {json_obj}, "
                f"reason: {key_error}")
            return None


class ImmediateMetadataSource(MetadataSource):

    TYPE = "ImmediateMetadataSource"

    def __init__(self, content: JSONObject):
        super().__init__()
        self.content = content

    def __eq__(self, other):
        return (
            isinstance(other, ImmediateMetadataSource)
            and self.content == other.content)

    def copy_object_to(self, destination: Path):
        pass

    def write_object_to(self, file_descriptor: IO):
        json.dump(self.content, file_descriptor)

    def deepcopy(self):
        return ImmediateMetadataSource(deepcopy(self.content))

    def to_json_obj(self) -> JSONObject:
        return {
            "@": dict(
                type="ImmediateMetadataSource",
                version=version_string
            ),
            MetadataSource.TYPE_KEY: ImmediateMetadataSource.TYPE,
            "content": self.content}

    @staticmethod
    def from_json_obj(json_obj: JSONObject) -> Optional["ImmediateMetadataSource"]:
        try:
            assert json_obj["@"]["type"] == "ImmediateMetadataSource"
            check_serialized_version(json_obj)
            assert json_obj[MetadataSource.TYPE_KEY] == ImmediateMetadataSource.TYPE
            return ImmediateMetadataSource(json_obj["content"])
        except KeyError as key_error:
            logger.error(
                f"could not read ImmediateMetadataSource from {json_obj}, "
                f"reason: {key_error}")
            return None



import tempfile
from collections import defaultdict
from pathlib import Path
from unittest.mock import patch

from nose.tools import (
    assert_equal,
    assert_in,
)

from ..utils import create_git_repo
from ..gitbackend.subprocess import (
    git_ls_tree,
    git_ls_tree_recursive,
)
from ..treeupdater import (
    EntryType,
    PathInfo,
    _get_dir,
    _write_dir,
    add_paths
)


default_repo = {
    "a": {
        "af1": "1234"
    },
    "b": {
        "bf1": "aasdsd"
    },
    "LICENSE": "license content"
}


def test_basic():
    path_infos = [
        PathInfo(["a", "b", "c"], "000000000000000000000000000000000000000c", EntryType.File),
        PathInfo(["a", "b", "d"], "000000000000000000000000000000000000000d", EntryType.File),
        PathInfo(["tools", "ttttc"], "000000000000000000000000000000000000000c", EntryType.File),
        PathInfo(["a", "d"], "000000000000000000000000000000000000000d", EntryType.File),
        PathInfo(["b", "a"], "000000000000000000000000000000000000000a", EntryType.File),
        PathInfo(["c", "b"], "000000000000000000000000000000000000000b", EntryType.File),
        PathInfo(["d"], "000000000000000000000000000000000000000d", EntryType.File),
        PathInfo(["LICENSE"], "00000000000000000000000000000000000000a0", EntryType.File)
    ]

    with tempfile.TemporaryDirectory() as td:
        repo_path = Path(td)
        tree_hash = create_git_repo(repo_path, default_repo)
        root_entries = _get_dir(repo_path, tree_hash)
        result = add_paths(repo_path, path_infos, root_entries)

        lines = git_ls_tree_recursive(str(repo_path), result)
        for path_info in path_infos:
            assert_in(
                f"100644 blob {path_info.object_hash}\t{'/'.join(path_info.elements)}",
                lines
            )


def test_dir_adding():
    path_infos = [
        PathInfo(
            ["a_dir"],
            "000000000000000000000000000000000000000c",
            EntryType.Directory
        )
    ]

    with tempfile.TemporaryDirectory() as td:
        repo_path = Path(td)
        tree_hash = create_git_repo(repo_path, default_repo)
        root_entries = _get_dir(repo_path, tree_hash)
        result = add_paths(repo_path, path_infos, root_entries)

        lines = git_ls_tree(str(repo_path), result)
        for path_info in path_infos:
            assert_in(
                f"040000 tree {path_info.object_hash}\t{'/'.join(path_info.elements)}",
                lines
            )


def test_dir_overwrite_error():
    path_infos = [
        PathInfo(
            ["a"],
            "000000000000000000000000000000000000000c",
            EntryType.File
        )
    ]

    with tempfile.TemporaryDirectory() as td:
        repo_path = Path(td)
        tree_hash = create_git_repo(repo_path, default_repo)
        root_entries = _get_dir(repo_path, tree_hash)

        try:
            add_paths(repo_path, path_infos, root_entries)
            raise RuntimeError("did not get expected ValueError")
        except ValueError as ve:
            assert_equal(ve.args[0], "cannot convert Directory to File: a")


def test_file_overwrite_error():
    path_infos = [
        PathInfo(
            ["LICENSE"],
            "000000000000000000000000000000000000000c",
            EntryType.Directory
        )
    ]

    with tempfile.TemporaryDirectory() as td:
        repo_path = Path(td)
        tree_hash = create_git_repo(repo_path, default_repo)
        root_entries = _get_dir(repo_path, tree_hash)

        try:
            add_paths(repo_path, path_infos, root_entries)
            raise RuntimeError("did not get expected ValueError")
        except ValueError as ve:
            assert_equal(
                ve.args[0],
                "cannot convert File to Directory: LICENSE")


def test_minimal_invocation():

    complex_repo = {
        "a": {
            "a_a": {
                "a_a_f1": "content of a_a_f1"
            },
            "a_b": {
                "a_b_f1": "content of a_b_f1"
            }
        },
        "b": {
            "b_a": {
                "b_a_f1": "content of b_a_f1"
            },
            "b_b": {
                "b_b_f1": "content of b_b_f1"
            }
        },
        "c": {
            "c_f1": "content of c_f1"
        }
    }

    path_infos = [
        PathInfo(
            ["a", "a_a", "a_a_a", "a_a_a_f1"],
            "000000000000000000000000000000000000000c",
            EntryType.File
        ),
        PathInfo(
            ["a", "a_a", "a_a_a", "a_a_a_f2"],
            "000000000000000000000000000000000000000c",
            EntryType.File
        ),
        PathInfo(
            ["a", "a_a", "a_a_f1"],
            "000000000000000000000000000000000000000c",
            EntryType.File
        ),
        PathInfo(
            ["a", "a_b", "a_b_f2"],
            "000000000000000000000000000000000000000c",
            EntryType.File
        ),
    ]

    with \
            tempfile.TemporaryDirectory() as td, \
            patch("dataladmetadatamodel.mapper.gitmapper.treeupdater._get_dir") as gd_mock, \
            patch("dataladmetadatamodel.mapper.gitmapper.treeupdater._write_dir") as wd_mock:

        def gd_counter(*args):
            get_calls[args[1]] += 1
            return _get_dir(*args)

        def wd_counter(*args):
            write_calls[args] += 1
            return _write_dir(*args)

        get_calls = defaultdict(int)
        write_calls = defaultdict(int)
        gd_mock.side_effect = gd_counter
        wd_mock.side_effect = wd_counter

        repo_path = Path(td)
        tree_hash = create_git_repo(repo_path, complex_repo)
        root_entries = _get_dir(repo_path, tree_hash)

        result = add_paths(repo_path, path_infos, root_entries)

        # Assert that a node is only read once
        assert all(map(lambda x: x == 1, get_calls.values()))
        assert len(get_calls) == 3
        # Assert that a node is only written once
        assert all(map(lambda x: x == 1, write_calls.values()))
        assert len(write_calls) == 5

        lines = git_ls_tree_recursive(str(repo_path), result)
        for path_info in path_infos:
            assert_in(
                f"100644 blob {path_info.object_hash}\t{'/'.join(path_info.elements)}",
                lines)



import unittest
from typing import Union

from ..mapper import Mapper
from ..reference import Reference
from ...datasettree import DatasetTree
from ...filetree import FileTree
from ...mappableobject import MappableObject


class DummyMapper(Mapper):
    def __init__(self,
                 class_name: str,
                 destination: str,
                 location: str):

        super().__init__(class_name, destination)
        self.location = location
        self.map_in_call_count = 0

    def map_in_impl(self,
                    mappable_object: MappableObject,
                    realm: str,
                    reference: Reference) -> None:
        self.map_in_call_count += 1
        return

    def map_out_impl(self,
                     mappable_object: MappableObject,
                     realm: str,
                     force_write: bool
                     ) -> Reference:
        return Reference(self.class_name,
                         self.location)


class TestMapper(unittest.TestCase):

    def _test_compatibility_tree(self,
                                tree_class: Union[type(DatasetTree), type(FileTree)]):
        mapper = DummyMapper("MTreeNode", "/tmp/1", "loc-1")
        tree = tree_class()
        mapper.map_in(tree,
                      "/tmp/1",
                      Reference(tree_class.__name__,
                                "location0"))
        self.assertEqual(mapper.map_in_call_count, 1)

    def test_compatibility_datasettree(self):
        self._test_compatibility_tree(DatasetTree)

    def test_compatibility_filetree(self):
        self._test_compatibility_tree(FileTree)


if __name__ == '__main__':
    unittest.main()



import unittest
from typing import Tuple

from dataladmetadatamodel.metadata import (
    ExtractorConfiguration,
    MetadataInstance,
    MetadataInstanceSet
)


class TestInstanceSetBase(unittest.TestCase):
    def setUp(self) -> None:

        self.metadata_instances = {
            prefix: self.create_metadata_instance(time_stamp, prefix)
            for time_stamp, prefix in ((1.0, "default"), (2.0, "new"))
        }

        self.metadata_instance_set = MetadataInstanceSet()
        self.metadata_instance_set.add_metadata_instance(
            self.metadata_instances["default"][1]
        )

    def create_metadata_instance(self,
                                 time_stamp: float,
                                 prefix: str
                                 ) -> Tuple[ExtractorConfiguration, MetadataInstance]:

        extractor_configuration = ExtractorConfiguration(
            "1.0",
            {"info": f"{prefix}_value"}
        )

        return (
            extractor_configuration,
            MetadataInstance(
                time_stamp,
                f"{prefix}_name",
                f"{prefix}_email",
                extractor_configuration,
                {"metadata 1": f"{prefix}_content"}
            )
        )

    def get_instance(self, prefix):
        return self.metadata_instances[prefix][1]

    def get_configuration(self, prefix):
        return self.metadata_instances[prefix][0]

    def assert_equal_to_pattern(self, instance_set: MetadataInstanceSet) -> None:
        self.assertListEqual(
            self.metadata_instance_set.parameter_set,
            instance_set.parameter_set
        )

        self.assertDictEqual(
            self.metadata_instance_set._instances,
            instance_set._instances
        )


class TestSerialization(TestInstanceSetBase):

    def test_object_round_trip(self):
        json_obj = self.metadata_instance_set.to_json_obj()
        ins2 = MetadataInstanceSet.from_json_obj(json_obj)
        self.assert_equal_to_pattern(ins2)

    def test_string_round_trip(self):
        json_str = self.metadata_instance_set.to_json_str()
        ins2 = MetadataInstanceSet.from_json_str(json_str)
        self.assert_equal_to_pattern(ins2)


class TestUniqueness(TestInstanceSetBase):

    def test_configuration_unity(self):
        # Ensure known state
        configuration_list = self.metadata_instance_set.configurations
        self.assertEqual(1, len(configuration_list))
        self.assertEqual(configuration_list[0], self.get_configuration("default"))

        # Re-add the only instance
        self.metadata_instance_set.add_metadata_instance(
            self.metadata_instance_set.get_instance_for_configuration(
                configuration_list[0]))

        # Ensure nothing has changes
        new_configuration_list = self.metadata_instance_set.configurations
        self.assertEqual(configuration_list, new_configuration_list)
        self.assertEqual(
            self.metadata_instance_set.get_instance_for_configuration(
                configuration_list[0]
            ),
            self.get_instance("default")
        )

    def test_configuration_multiplicity(self):
        configuration_list = self.metadata_instance_set.configurations
        self.assertEqual(1, len(configuration_list))
        self.assertEqual(configuration_list[0], self.get_configuration("default"))

        self.metadata_instance_set.add_metadata_instance(self.get_instance("new"))
        configuration_list = self.metadata_instance_set.configurations
        self.assertEqual(2, len(configuration_list))
        self.assertIn(self.get_configuration("default"), configuration_list)
        self.assertIn(self.get_configuration("new"), configuration_list)


if __name__ == '__main__':
    unittest.main()



"""
Instances of the Text class just contain
text. Their main use is as dummy-element
during model development.
"""
from typing import (
    Iterable,
    Optional
)

from dataladmetadatamodel.mappableobject import MappableObject
from dataladmetadatamodel.mapper.reference import Reference


class Text(MappableObject):

    def __init__(self,
                 content: Optional[str] = None,
                 realm: Optional[str] = None,
                 reference: Optional[Reference] = None):

        assert isinstance(content, (type(None), str))
        assert isinstance(reference, (type(None), Reference))

        super().__init__(realm, reference)
        self.content = content

    @staticmethod
    def get_empty_instance(realm: Optional[str] = None,
                           reference: Optional[Reference] = None):
        return Text(None, realm, reference)

    def modifiable_sub_objects_impl(self) -> Iterable[MappableObject]:
        return []

    def purge_impl(self):
        self.content = None

    def deepcopy_impl(self,
                      new_mapper_family: Optional[str] = None,
                      new_destination: Optional[str] = None,
                      **kwargs) -> "Text":

        copied_text = Text()
        copied_text.write_out(new_destination)
        return copied_text



from typing import (
    Dict,
    Iterable,
    Optional
)

from uuid import UUID

from dataladmetadatamodel.mappableobject import MappableObject
from dataladmetadatamodel.versionlist import VersionList
from dataladmetadatamodel.mapper.reference import Reference


class UUIDSet(MappableObject):
    def __init__(self,
                 initial_set: Optional[Dict[UUID, VersionList]] = None,
                 realm: Optional[str] = None,
                 reference: Optional[Reference] = None):

        assert isinstance(initial_set, (type(None), dict))
        assert isinstance(realm, (type(None), str))
        assert isinstance(reference, (type(None), Reference))

        super().__init__(realm, reference)
        self.uuid_set = initial_set or dict()

    def modifiable_sub_objects_impl(self) -> Iterable[MappableObject]:
        return self.uuid_set.values()

    def purge_impl(self):
        for version_list in self.uuid_set.values():
            version_list.purge()
        self.uuid_set = dict()

    def uuids(self):
        self.ensure_mapped()
        return self.uuid_set.keys()

    def set_version_list(self,
                         uuid: UUID,
                         version_list: VersionList):
        """
        Set a new or updated version list for the uuid.
        Existing references are deleted.
        The entry is marked as dirty.
        """
        self.touch()
        self.uuid_set[uuid] = version_list

    def get_version_list(self, uuid: UUID) -> VersionList:
        """
        Get the version list for uuid. If it is not mapped yet,
        it will be mapped.
        """
        self.ensure_mapped()
        return self.uuid_set[uuid].read_in()

    def unget_version_list(self, uuid):
        """
        Remove a version list from memory. First, persist the
        current status, if it was changed.
        """
        self.uuid_set[uuid].write_out()
        self.uuid_set[uuid].purge()

    def deepcopy_impl(self,
                      new_mapper_family: Optional[str] = None,
                      new_destination: Optional[str] = None,
                      **kwargs) -> "UUIDSet":

        copied_uuid_set = UUIDSet()
        for uuid, version_list in self.uuid_set.items():
            copied_uuid_set.uuid_set[uuid] = version_list.deepcopy(
                new_mapper_family,
                new_destination)

        copied_uuid_set.write_out(new_destination)
        copied_uuid_set.purge()
        return copied_uuid_set



from typing import (
    Dict,
    List,
    Union
)


JSONObject = Union[
    None, bool, int, float, str,
    List["JSONObject"],
    Dict[str, "JSONObject"]
]


version = 2
version_minor = 0
version_string = f"{version}.{version_minor}"


def check_serialized_version(json_object: JSONObject):
    stored_class = json_object["@"]["type"]
    stored_version = json_object["@"]["version"]
    if stored_version != version_string:
        raise ValueError(
            f"Unsupported metadata version ({stored_version}) in "
            f"stored {stored_class} object, expected version: "
            f"{version_string}")


from . import _version
__version__ = _version.get_versions()['version']



import subprocess
import tempfile
import unittest
from pathlib import Path

from dataladmetadatamodel.metadatapath import MetadataPath
from dataladmetadatamodel.versionlist import (
    VersionList,
    VersionRecord
)
from dataladmetadatamodel.mapper.gitmapper.objectreference import flush_object_references

from .utils import (
    assert_dataset_trees_equal,
    create_dataset_tree
)


file_test_paths = [
    MetadataPath("a/b/c"),
    MetadataPath("a/b/a"),
    MetadataPath("b"),
    MetadataPath("c/d/e"),
    MetadataPath("a/x")]

dataset_test_paths = [
    [
        MetadataPath(""),
        MetadataPath("d1"),
        MetadataPath("d1/d1.1"),
        MetadataPath("d2"),
        MetadataPath("d2/d2.1/d2.1.1"),
        MetadataPath("d3/d3.1")
    ],
    [
        MetadataPath(""),
        MetadataPath("e1"),
        MetadataPath("e1/e1.1"),
        MetadataPath("e2"),
        MetadataPath("e2/e2.1/e2.1.1"),
        MetadataPath("e3/e3.1")
    ]
]


class TestVersionList(unittest.TestCase):

    def test_basic(self):
        VersionList()

    def test_copy_end_to_end(self):
        with \
                tempfile.TemporaryDirectory() as original_dir, \
                tempfile.TemporaryDirectory() as copy_dir:

            subprocess.run(["git", "init", original_dir])
            subprocess.run(["git", "init", copy_dir])

            dataset_trees = [
                create_dataset_tree(
                    dataset_test_paths[index],
                    file_test_paths)
                for index in range(2)]

            dataset_trees[0].write_out(original_dir)
            dataset_trees[0].purge()

            dataset_trees[1].write_out(original_dir)

            version_list = VersionList(initial_set={
                "v1": {
                    MetadataPath("a"): VersionRecord(
                        "0.1",
                        MetadataPath("a"),
                        dataset_trees[0]
                    )
                },
                "v2": {
                    MetadataPath("b"): VersionRecord(
                        "0.2",
                        MetadataPath("b"),
                        dataset_trees[1]
                    )
                }
            })

            version_list.write_out(original_dir)

            path_prefix = "x/y/z"
            copied_version_list = version_list.deepcopy(
                new_destination=copy_dir,
                path_prefix=MetadataPath(path_prefix))

            # Check that the mapping state of the original
            # has not changed.
            self.assertFalse(dataset_trees[0].mtree.mapped)
            self.assertTrue(dataset_trees[1].mtree.mapped)

            # Read the copied version list
            copied_version_list.read_in()

            # Check that the copied elements are unmapped
            for _, (_, _, copied_dataset_tree) in copied_version_list.versioned_elements:
                self.assertFalse(copied_dataset_tree.mtree.mapped)

            # Compare the version lists, taking modified dataset tree paths into account
            for primary_version, (_, path, dataset_tree) in version_list.versioned_elements:
                expected_path = path_prefix / path
                _, copy_path, copied_dataset_tree = copied_version_list.get_versioned_element(primary_version, expected_path)
                self.assertEqual(expected_path, copy_path)
                dataset_tree.read_in()
                copied_dataset_tree.read_in()
                assert_dataset_trees_equal(self, dataset_tree, copied_dataset_tree, True)
                # NB! dataset tree copying is tested in dataset-tree related unit tests


if __name__ == '__main__':
    unittest.main()



from pathlib import PurePosixPath

from dataladmetadatamodel.log import logger


class MetadataPath(PurePosixPath):
    def __new__(cls, *args):

        posix_args = []
        for arg in args:
            if isinstance(arg, str):
                arg.replace("\\", "/")
            posix_args.append(arg)

        original_path = PurePosixPath(*posix_args)
        if not original_path.is_absolute():
            created_path = super().__new__(
                cls,
                "/".join(original_path.parts))
        else:
            created_path = super().__new__(
                cls,
                "/".join(original_path.parts[1:]))
            logger.warning(
                f"Denied creation of absolute metadata path: {original_path}, "
                f"created {created_path} instead. This is considered an error "
                f"in the calling code.")

        return created_path

    def __str__(self):
        path_str = PurePosixPath.__str__(self)
        return (
            ""
            if path_str == "."
            else path_str)

    def __len__(self):
        return len(self.parts)



from uuid import UUID

from dataladmetadatamodel.mapper.gitmapper.gitbackend.subprocess import (
    git_load_json,
    git_save_json,
)
from dataladmetadatamodel.mapper.mapper import Mapper
from dataladmetadatamodel.mapper.reference import Reference


class Strings:
    GIT = "git"
    DATASET_IDENTIFIER = "dataset_identifier"
    DATASET_VERSION = "dataset_version"
    DATASET_LEVEL_METADATA = "dataset_level_metadata"
    FILE_TREE = "file_tree"


class MetadataRootRecordGitMapper(Mapper):
    def map_in_impl(self,
                    metadata_root_record: "MetadataRootRecord",
                    realm: str,
                    reference: Reference) -> None:

        from dataladmetadatamodel.filetree import FileTree
        from dataladmetadatamodel.metadata import Metadata
        from dataladmetadatamodel.metadatarootrecord import MetadataRootRecord

        assert isinstance(metadata_root_record, MetadataRootRecord)
        assert isinstance(realm, str)
        assert isinstance(reference, Reference)

        json_object = git_load_json(realm, reference.location)

        metadata_reference = Reference.from_json_obj(
            json_object[Strings.DATASET_LEVEL_METADATA])
        if metadata_reference.is_none_reference():
            metadata = None
        else:
            metadata = Metadata(realm=realm, reference=metadata_reference)

        file_tree_reference = Reference.from_json_obj(
            json_object[Strings.FILE_TREE])
        if file_tree_reference.is_none_reference():
            file_tree = None
        else:
            file_tree = FileTree(realm=realm, reference=file_tree_reference)

        MetadataRootRecord.__init__(
            metadata_root_record,
            UUID(json_object[Strings.DATASET_IDENTIFIER]),
            json_object[Strings.DATASET_VERSION],
            metadata,
            file_tree,
            realm=realm,
            reference=reference)

    def map_out_impl(self,
                     mrr: "MetadataRootRecord",
                     realm: str,
                     force_write: bool) -> Reference:

        from dataladmetadatamodel.metadatarootrecord import MetadataRootRecord

        assert isinstance(mrr, MetadataRootRecord)

        if mrr._file_tree is None:
            file_tree_reference = Reference.get_none_reference("FileTree")
        else:
            file_tree_reference = mrr._file_tree.write_out(
                realm,
                "git",
                force_write)

        if mrr.dataset_level_metadata is None:
            dataset_level_metadata_reference = Reference.get_none_reference("Metadata")
        else:
            dataset_level_metadata_reference = mrr.dataset_level_metadata.write_out(
                realm,
                "git",
                force_write)

        json_object = {
            Strings.DATASET_IDENTIFIER: str(mrr.dataset_identifier),
            Strings.DATASET_VERSION: str(mrr.dataset_version),
            Strings.DATASET_LEVEL_METADATA: dataset_level_metadata_reference.to_json_obj(),
            Strings.FILE_TREE: file_tree_reference.to_json_obj()}

        return Reference(
            "MetadataRootRecord",
            git_save_json(realm, json_object))
